/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://rtsys.informatik.uni-kiel.de/kieler
 * 
 * Copyright 2015 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 * 
 * SPDX-License-Identifier: EPL-2.0
 */
package de.cau.cs.kieler.klighd.ide.syntheses;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.Iterables;
import com.google.common.collect.Table;
import de.cau.cs.kieler.klighd.KlighdConstants;
import de.cau.cs.kieler.klighd.SynthesisOption;
import de.cau.cs.kieler.klighd.ide.syntheses.action.EcoreModelExpandDetailsAction;
import de.cau.cs.kieler.klighd.kgraph.KEdge;
import de.cau.cs.kieler.klighd.kgraph.KLabel;
import de.cau.cs.kieler.klighd.kgraph.KNode;
import de.cau.cs.kieler.klighd.kgraph.KPort;
import de.cau.cs.kieler.klighd.kgraph.util.KGraphUtil;
import de.cau.cs.kieler.klighd.krendering.Colors;
import de.cau.cs.kieler.klighd.krendering.KAreaPlacementData;
import de.cau.cs.kieler.klighd.krendering.KPolyline;
import de.cau.cs.kieler.klighd.krendering.KRectangle;
import de.cau.cs.kieler.klighd.krendering.KRoundedRectangle;
import de.cau.cs.kieler.klighd.krendering.KText;
import de.cau.cs.kieler.klighd.krendering.LineStyle;
import de.cau.cs.kieler.klighd.krendering.extensions.KColorExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KContainerRenderingExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KEdgeExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KLabelExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KNodeExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KPolylineExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KPortExtensions;
import de.cau.cs.kieler.klighd.krendering.extensions.KRenderingExtensions;
import de.cau.cs.kieler.klighd.microlayout.Bounds;
import de.cau.cs.kieler.klighd.microlayout.PlacementUtil;
import de.cau.cs.kieler.klighd.syntheses.AbstractDiagramSynthesis;
import de.cau.cs.kieler.klighd.syntheses.DiagramSyntheses;
import de.cau.cs.kieler.klighd.util.KlighdProperties;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Consumer;
import javax.inject.Inject;
import org.eclipse.elk.alg.layered.options.EdgeStraighteningStrategy;
import org.eclipse.elk.alg.layered.options.FixedAlignment;
import org.eclipse.elk.alg.layered.options.LayeredOptions;
import org.eclipse.elk.core.options.CoreOptions;
import org.eclipse.elk.core.options.Direction;
import org.eclipse.elk.core.options.PortSide;
import org.eclipse.elk.core.options.SizeConstraint;
import org.eclipse.elk.core.util.Pair;
import org.eclipse.elk.graph.properties.IProperty;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.util.EContentsEList;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.IteratorExtensions;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

/**
 * Diagram synthesis of a {@link EObject}.
 * 
 * @author als
 * @kieler.design 2014-07-30 proposed
 * @kieler.rating 2014-07-30 proposed yellow
 */
@SuppressWarnings("all")
public class EObjectFallbackSynthesis extends AbstractDiagramSynthesis<EObject> {
  @Inject
  @Extension
  private KNodeExtensions _kNodeExtensions;
  
  @Inject
  @Extension
  private KEdgeExtensions _kEdgeExtensions;
  
  @Inject
  @Extension
  private KPortExtensions _kPortExtensions;
  
  @Inject
  @Extension
  private KLabelExtensions _kLabelExtensions;
  
  @Inject
  @Extension
  private KRenderingExtensions _kRenderingExtensions;
  
  @Inject
  @Extension
  private KContainerRenderingExtensions _kContainerRenderingExtensions;
  
  @Inject
  @Extension
  private KPolylineExtensions _kPolylineExtensions;
  
  @Inject
  @Extension
  private KColorExtensions _kColorExtensions;
  
  public static final String ID = "de.cau.cs.kieler.klighd.ui.view.syntheses.EObjectFallbackSynthesis";
  
  public static final SynthesisOption EXPAND_DETAILS = SynthesisOption.createCheckOption(EObjectFallbackSynthesis.class, "Expand all Details", 
    Boolean.valueOf(false)).setUpdateAction(EcoreModelExpandDetailsAction.ID);
  
  public static final SynthesisOption SHOW_REFERENCES = SynthesisOption.createCheckOption(EObjectFallbackSynthesis.class, "Show References", Boolean.valueOf(false));
  
  @Override
  public List<SynthesisOption> getDisplayedSynthesisOptions() {
    return CollectionLiterals.<SynthesisOption>newLinkedList(EObjectFallbackSynthesis.EXPAND_DETAILS, EObjectFallbackSynthesis.SHOW_REFERENCES);
  }
  
  @Override
  public List<Pair<IProperty<?>, List<?>>> getDisplayedLayoutOptions() {
    final Function1<Direction, String> _function = new Function1<Direction, String>() {
      @Override
      public String apply(final Direction it) {
        return it.name();
      }
    };
    List<Direction> _sortBy = IterableExtensions.<Direction, String>sortBy(IterableExtensions.<Direction>drop(((Iterable<Direction>)Conversions.doWrapArray(Direction.values())), 1), _function);
    Pair<IProperty<?>, List<?>> _pair = new Pair<IProperty<?>, List<?>>(CoreOptions.DIRECTION, _sortBy);
    ArrayList<Integer> _newArrayList = CollectionLiterals.<Integer>newArrayList(Integer.valueOf(0), Integer.valueOf(150));
    Pair<IProperty<?>, List<?>> _pair_1 = new Pair<IProperty<?>, List<?>>(CoreOptions.SPACING_NODE_NODE, _newArrayList);
    return CollectionLiterals.<Pair<IProperty<?>, List<?>>>newLinkedList(_pair, _pair_1);
  }
  
  private final Map<EObject, KNode> nodeCache = CollectionLiterals.<EObject, KNode>newHashMap();
  
  private final Table<EStructuralFeature, KNode, KPort> portCache = HashBasedTable.<EStructuralFeature, KNode, KPort>create();
  
  @Override
  public KNode transform(final EObject model) {
    this.nodeCache.clear();
    this.portCache.clear();
    final KNode rootNode = this._kNodeExtensions.createNode();
    this._kNodeExtensions.<String>addLayoutParam(rootNode, CoreOptions.ALGORITHM, LayeredOptions.ALGORITHM_ID);
    this.<KNode, FixedAlignment>setLayoutOption(rootNode, LayeredOptions.NODE_PLACEMENT_BK_FIXED_ALIGNMENT, FixedAlignment.BALANCED);
    this.<KNode, EdgeStraighteningStrategy>setLayoutOption(rootNode, LayeredOptions.NODE_PLACEMENT_BK_EDGE_STRAIGHTENING, EdgeStraighteningStrategy.IMPROVE_STRAIGHTNESS);
    Double _default = LayeredOptions.SPACING_NODE_NODE.getDefault();
    double _multiply = ((_default).doubleValue() * 1.1f);
    this.<KNode, Double>setLayoutOption(rootNode, LayeredOptions.SPACING_EDGE_NODE, Double.valueOf(_multiply));
    Double _default_1 = LayeredOptions.SPACING_NODE_NODE.getDefault();
    double _multiply_1 = ((_default_1).doubleValue() * 1.1f);
    this.<KNode, Double>setLayoutOption(rootNode, LayeredOptions.SPACING_EDGE_NODE_BETWEEN_LAYERS, Double.valueOf(_multiply_1));
    EList<KNode> _children = rootNode.getChildren();
    KNode _transformToNode = this.transformToNode(model);
    _children.add(_transformToNode);
    EList<KNode> _children_1 = rootNode.getChildren();
    final Function1<EObject, KNode> _function = new Function1<EObject, KNode>() {
      @Override
      public KNode apply(final EObject it) {
        return EObjectFallbackSynthesis.this.transformToNode(it);
      }
    };
    Iterable<KNode> _iterable = IteratorExtensions.<KNode>toIterable(IteratorExtensions.<EObject, KNode>map(model.eAllContents(), _function));
    Iterables.<KNode>addAll(_children_1, _iterable);
    Set<EObject> _keySet = this.nodeCache.keySet();
    for (final EObject eObject : _keySet) {
      {
        final EObject eContainer = eObject.eContainer();
        if ((eContainer != null)) {
          KEdge _createEdge = this._kEdgeExtensions.createEdge();
          final Procedure1<KEdge> _function_1 = new Procedure1<KEdge>() {
            @Override
            public void apply(final KEdge it) {
              final KNode sourceNode = EObjectFallbackSynthesis.this.nodeCache.get(eContainer);
              it.setSource(sourceNode);
              final EStructuralFeature eContainingFeature = eObject.eContainingFeature();
              boolean _contains = EObjectFallbackSynthesis.this.portCache.contains(eContainingFeature, sourceNode);
              if (_contains) {
                it.setSourcePort(EObjectFallbackSynthesis.this.portCache.get(eContainingFeature, sourceNode));
              } else {
                it.setSourcePort(EObjectFallbackSynthesis.this.transformToPort(eContainingFeature, sourceNode));
              }
              it.setTarget(EObjectFallbackSynthesis.this.nodeCache.get(eObject));
              KPolyline _addPolyline = EObjectFallbackSynthesis.this._kEdgeExtensions.addPolyline(it);
              final Procedure1<KPolyline> _function = new Procedure1<KPolyline>() {
                @Override
                public void apply(final KPolyline it) {
                  EObjectFallbackSynthesis.this._kPolylineExtensions.addHeadArrowDecorator(it);
                  EObjectFallbackSynthesis.this._kPolylineExtensions.addJunctionPointDecorator(it);
                }
              };
              ObjectExtensions.<KPolyline>operator_doubleArrow(_addPolyline, _function);
            }
          };
          ObjectExtensions.<KEdge>operator_doubleArrow(_createEdge, _function_1);
        }
        boolean _booleanValue = this.getBooleanValue(EObjectFallbackSynthesis.SHOW_REFERENCES);
        if (_booleanValue) {
          for (final EContentsEList.FeatureIterator<EObject> featureIterator = ((EContentsEList.FeatureIterator<EObject>) eObject.eCrossReferences().iterator()); featureIterator.hasNext();) {
            {
              EObject _next = featureIterator.next();
              final EObject targetEObject = ((EObject) _next);
              EStructuralFeature _feature = featureIterator.feature();
              final EReference eReference = ((EReference) _feature);
              KEdge _createEdge_1 = this._kEdgeExtensions.createEdge();
              final Procedure1<KEdge> _function_2 = new Procedure1<KEdge>() {
                @Override
                public void apply(final KEdge it) {
                  final KNode sourceNode = EObjectFallbackSynthesis.this.nodeCache.get(eObject);
                  it.setSource(sourceNode);
                  boolean _contains = EObjectFallbackSynthesis.this.portCache.contains(eReference, sourceNode);
                  if (_contains) {
                    it.setSourcePort(EObjectFallbackSynthesis.this.portCache.get(eReference, sourceNode));
                  } else {
                    it.setSourcePort(EObjectFallbackSynthesis.this.transformToPort(eReference, sourceNode));
                  }
                  it.setTarget(EObjectFallbackSynthesis.this.nodeCache.get(targetEObject));
                  KPolyline _addPolyline = EObjectFallbackSynthesis.this._kEdgeExtensions.addPolyline(it);
                  final Procedure1<KPolyline> _function = new Procedure1<KPolyline>() {
                    @Override
                    public void apply(final KPolyline it) {
                      EObjectFallbackSynthesis.this._kPolylineExtensions.addHeadArrowDecorator(it);
                      EObjectFallbackSynthesis.this._kPolylineExtensions.addJunctionPointDecorator(it);
                      EObjectFallbackSynthesis.this._kRenderingExtensions.setLineStyle(it, LineStyle.DASH);
                    }
                  };
                  ObjectExtensions.<KPolyline>operator_doubleArrow(_addPolyline, _function);
                }
              };
              ObjectExtensions.<KEdge>operator_doubleArrow(_createEdge_1, _function_2);
            }
          }
        }
      }
    }
    return rootNode;
  }
  
  /**
   * Translate a single EObject with super type and attributes into a node.
   */
  private KNode transformToNode(final EObject object) {
    final KNode node = this._kNodeExtensions.createNode(object);
    this.<KNode>associateWith(node, object);
    this.<KNode, Boolean>setLayoutOption(node, KlighdProperties.EXPAND, Boolean.valueOf(this.getBooleanValue(EObjectFallbackSynthesis.EXPAND_DETAILS)));
    this.<KNode, EnumSet<SizeConstraint>>setLayoutOption(node, CoreOptions.NODE_SIZE_CONSTRAINTS, SizeConstraint.free());
    final EClass eclass = object.eClass();
    boolean _isEmpty = eclass.getEAllAttributes().isEmpty();
    final boolean hasAttributes = (!_isEmpty);
    boolean _isEmpty_1 = eclass.getEAllSuperTypes().isEmpty();
    final boolean hasSuperTypes = (!_isEmpty_1);
    KRoundedRectangle _createFigure = this.createFigure(node);
    final Procedure1<KRoundedRectangle> _function = new Procedure1<KRoundedRectangle>() {
      @Override
      public void apply(final KRoundedRectangle it) {
        it.<Boolean>setProperty(KlighdProperties.EXPANDED_RENDERING, Boolean.valueOf(true));
        EObjectFallbackSynthesis.this._kContainerRenderingExtensions.setGridPlacement(it, 1);
        KText _associateWith = EObjectFallbackSynthesis.this.<KText>associateWith(EObjectFallbackSynthesis.this._kContainerRenderingExtensions.addText(it, object.eClass().getName()), object);
        final Procedure1<KText> _function = new Procedure1<KText>() {
          @Override
          public void apply(final KText it) {
            EObjectFallbackSynthesis.this._kRenderingExtensions.setFontSize(it, 11);
            EObjectFallbackSynthesis.this._kRenderingExtensions.<KText>setFontBold(it, true);
            EObjectFallbackSynthesis.this._kRenderingExtensions.to(EObjectFallbackSynthesis.this._kRenderingExtensions.from(EObjectFallbackSynthesis.this._kRenderingExtensions.setGridPlacementData(it), EObjectFallbackSynthesis.this._kRenderingExtensions.LEFT, 9, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.TOP, 8f, 0), EObjectFallbackSynthesis.this._kRenderingExtensions.RIGHT, 8, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.BOTTOM, 4, 0);
            DiagramSyntheses.suppressSelectability(it);
          }
        };
        ObjectExtensions.<KText>operator_doubleArrow(_associateWith, _function);
        KText _addText = EObjectFallbackSynthesis.this._kContainerRenderingExtensions.addText(it, "[Hide]");
        final Procedure1<KText> _function_1 = new Procedure1<KText>() {
          @Override
          public void apply(final KText it) {
            EObjectFallbackSynthesis.this._kRenderingExtensions.setForeground(it, EObjectFallbackSynthesis.this._kColorExtensions.getColor("blue"));
            EObjectFallbackSynthesis.this._kRenderingExtensions.setFontSize(it, 9);
            EObjectFallbackSynthesis.this._kRenderingExtensions.<KText>addSingleClickAction(it, KlighdConstants.ACTION_COLLAPSE_EXPAND);
            EObjectFallbackSynthesis.this._kRenderingExtensions.<KText>addDoubleClickAction(it, KlighdConstants.ACTION_COLLAPSE_EXPAND);
            EObjectFallbackSynthesis.this._kRenderingExtensions.to(EObjectFallbackSynthesis.this._kRenderingExtensions.from(EObjectFallbackSynthesis.this._kRenderingExtensions.setGridPlacementData(it), EObjectFallbackSynthesis.this._kRenderingExtensions.LEFT, 8, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.TOP, 0, 0), EObjectFallbackSynthesis.this._kRenderingExtensions.RIGHT, 8, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.BOTTOM, 0, 0);
          }
        };
        ObjectExtensions.<KText>operator_doubleArrow(_addText, _function_1);
        if ((hasAttributes || hasSuperTypes)) {
          KRectangle _addRectangle = EObjectFallbackSynthesis.this._kContainerRenderingExtensions.addRectangle(it);
          final Procedure1<KRectangle> _function_2 = new Procedure1<KRectangle>() {
            @Override
            public void apply(final KRectangle it) {
              EObjectFallbackSynthesis.this._kRenderingExtensions.to(EObjectFallbackSynthesis.this._kRenderingExtensions.from(EObjectFallbackSynthesis.this._kRenderingExtensions.setGridPlacementData(it), EObjectFallbackSynthesis.this._kRenderingExtensions.LEFT, 8, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.TOP, 0, 0), EObjectFallbackSynthesis.this._kRenderingExtensions.RIGHT, 8, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.BOTTOM, 8, 0);
              EObjectFallbackSynthesis.this._kRenderingExtensions.setBackground(it, EObjectFallbackSynthesis.this._kColorExtensions.getColor("white"));
              EObjectFallbackSynthesis.this._kRenderingExtensions.setForeground(it, EObjectFallbackSynthesis.this._kColorExtensions.getColor("gray"));
              EObjectFallbackSynthesis.this._kRenderingExtensions.setLineWidth(it, 1);
              EObjectFallbackSynthesis.this._kContainerRenderingExtensions.setGridPlacement(it, 1);
              final KRectangle container = it;
              if (hasSuperTypes) {
                KText _addText = EObjectFallbackSynthesis.this._kContainerRenderingExtensions.addText(it, "SuperTypes");
                final Procedure1<KText> _function = new Procedure1<KText>() {
                  @Override
                  public void apply(final KText it) {
                    EObjectFallbackSynthesis.this._kRenderingExtensions.setForeground(it, EObjectFallbackSynthesis.this._kColorExtensions.getColor("gray"));
                    EObjectFallbackSynthesis.this._kRenderingExtensions.setFontSize(it, 8);
                    EObjectFallbackSynthesis.this._kRenderingExtensions.to(EObjectFallbackSynthesis.this._kRenderingExtensions.from(EObjectFallbackSynthesis.this._kRenderingExtensions.setGridPlacementData(it), EObjectFallbackSynthesis.this._kRenderingExtensions.LEFT, 5, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.TOP, 2, 0), EObjectFallbackSynthesis.this._kRenderingExtensions.RIGHT, 5, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.BOTTOM, 2, 0);
                    DiagramSyntheses.suppressSelectability(it);
                  }
                };
                ObjectExtensions.<KText>operator_doubleArrow(_addText, _function);
                final Consumer<EClass> _function_1 = new Consumer<EClass>() {
                  @Override
                  public void accept(final EClass it) {
                    KText _addText = EObjectFallbackSynthesis.this._kContainerRenderingExtensions.addText(container, it.getName());
                    final Procedure1<KText> _function = new Procedure1<KText>() {
                      @Override
                      public void apply(final KText it) {
                        EObjectFallbackSynthesis.this._kRenderingExtensions.setFontSize(it, 9);
                        EObjectFallbackSynthesis.this._kRenderingExtensions.to(EObjectFallbackSynthesis.this._kRenderingExtensions.from(EObjectFallbackSynthesis.this._kRenderingExtensions.setGridPlacementData(it), EObjectFallbackSynthesis.this._kRenderingExtensions.LEFT, 5, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.TOP, 2, 0), EObjectFallbackSynthesis.this._kRenderingExtensions.RIGHT, 5, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.BOTTOM, 2, 0);
                        DiagramSyntheses.suppressSelectability(it);
                      }
                    };
                    ObjectExtensions.<KText>operator_doubleArrow(_addText, _function);
                  }
                };
                IterableExtensions.<EClass>filterNull(eclass.getEAllSuperTypes()).forEach(_function_1);
              }
              if (hasAttributes) {
                if (hasSuperTypes) {
                  KPolyline _addHorizontalSeperatorLine = EObjectFallbackSynthesis.this._kContainerRenderingExtensions.addHorizontalSeperatorLine(it, 1, 1);
                  EObjectFallbackSynthesis.this._kRenderingExtensions.setForeground(_addHorizontalSeperatorLine, EObjectFallbackSynthesis.this._kColorExtensions.getColor("gray"));
                }
                KText _addText_1 = EObjectFallbackSynthesis.this._kContainerRenderingExtensions.addText(it, "Attributes");
                final Procedure1<KText> _function_2 = new Procedure1<KText>() {
                  @Override
                  public void apply(final KText it) {
                    EObjectFallbackSynthesis.this._kRenderingExtensions.setForeground(it, EObjectFallbackSynthesis.this._kColorExtensions.getColor("gray"));
                    EObjectFallbackSynthesis.this._kRenderingExtensions.setFontSize(it, 8);
                    EObjectFallbackSynthesis.this._kRenderingExtensions.to(EObjectFallbackSynthesis.this._kRenderingExtensions.from(EObjectFallbackSynthesis.this._kRenderingExtensions.setGridPlacementData(it), EObjectFallbackSynthesis.this._kRenderingExtensions.LEFT, 5, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.TOP, 2, 0), EObjectFallbackSynthesis.this._kRenderingExtensions.RIGHT, 5, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.BOTTOM, 2, 0);
                    DiagramSyntheses.suppressSelectability(it);
                  }
                };
                ObjectExtensions.<KText>operator_doubleArrow(_addText_1, _function_2);
                final Consumer<EAttribute> _function_3 = new Consumer<EAttribute>() {
                  @Override
                  public void accept(final EAttribute it) {
                    String _name = it.getName();
                    String _plus = (_name + ": ");
                    String _valueOf = String.valueOf(object.eGet(it));
                    String _plus_1 = (_plus + _valueOf);
                    KText _addText = EObjectFallbackSynthesis.this._kContainerRenderingExtensions.addText(container, _plus_1);
                    final Procedure1<KText> _function = new Procedure1<KText>() {
                      @Override
                      public void apply(final KText it) {
                        EObjectFallbackSynthesis.this._kRenderingExtensions.setFontSize(it, 9);
                        EObjectFallbackSynthesis.this._kRenderingExtensions.to(EObjectFallbackSynthesis.this._kRenderingExtensions.from(EObjectFallbackSynthesis.this._kRenderingExtensions.setGridPlacementData(it), EObjectFallbackSynthesis.this._kRenderingExtensions.LEFT, 5, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.TOP, 2, 0), EObjectFallbackSynthesis.this._kRenderingExtensions.RIGHT, 5, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.BOTTOM, 2, 0);
                        DiagramSyntheses.suppressSelectability(it);
                      }
                    };
                    ObjectExtensions.<KText>operator_doubleArrow(_addText, _function);
                  }
                };
                IterableExtensions.<EAttribute>filterNull(eclass.getEAllAttributes()).forEach(_function_3);
              }
            }
          };
          ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle, _function_2);
        }
      }
    };
    ObjectExtensions.<KRoundedRectangle>operator_doubleArrow(_createFigure, _function);
    KRoundedRectangle _createFigure_1 = this.createFigure(node);
    final Procedure1<KRoundedRectangle> _function_1 = new Procedure1<KRoundedRectangle>() {
      @Override
      public void apply(final KRoundedRectangle it) {
        it.<Boolean>setProperty(KlighdProperties.COLLAPSED_RENDERING, Boolean.valueOf(true));
        EObjectFallbackSynthesis.this._kContainerRenderingExtensions.setGridPlacement(it, 1);
        KText _associateWith = EObjectFallbackSynthesis.this.<KText>associateWith(EObjectFallbackSynthesis.this._kContainerRenderingExtensions.addText(it, object.eClass().getName()), object);
        final Procedure1<KText> _function = new Procedure1<KText>() {
          @Override
          public void apply(final KText it) {
            EObjectFallbackSynthesis.this._kRenderingExtensions.setFontSize(it, 11);
            EObjectFallbackSynthesis.this._kRenderingExtensions.<KText>setFontBold(it, true);
            KAreaPlacementData _from = EObjectFallbackSynthesis.this._kRenderingExtensions.from(EObjectFallbackSynthesis.this._kRenderingExtensions.setGridPlacementData(it), EObjectFallbackSynthesis.this._kRenderingExtensions.LEFT, 8, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.TOP, 8, 0);
            int _xifexpression = (int) 0;
            if ((hasAttributes || hasSuperTypes)) {
              _xifexpression = 4;
            } else {
              _xifexpression = 8;
            }
            EObjectFallbackSynthesis.this._kRenderingExtensions.to(_from, EObjectFallbackSynthesis.this._kRenderingExtensions.RIGHT, 8, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.BOTTOM, _xifexpression, 0);
            DiagramSyntheses.suppressSelectability(it);
          }
        };
        ObjectExtensions.<KText>operator_doubleArrow(_associateWith, _function);
        if ((hasAttributes || hasSuperTypes)) {
          KText _addText = EObjectFallbackSynthesis.this._kContainerRenderingExtensions.addText(it, "[Details]");
          final Procedure1<KText> _function_1 = new Procedure1<KText>() {
            @Override
            public void apply(final KText it) {
              EObjectFallbackSynthesis.this._kRenderingExtensions.setForeground(it, EObjectFallbackSynthesis.this._kColorExtensions.getColor("blue"));
              EObjectFallbackSynthesis.this._kRenderingExtensions.setFontSize(it, 9);
              EObjectFallbackSynthesis.this._kRenderingExtensions.<KText>addSingleClickAction(it, KlighdConstants.ACTION_COLLAPSE_EXPAND);
              EObjectFallbackSynthesis.this._kRenderingExtensions.<KText>addDoubleClickAction(it, KlighdConstants.ACTION_COLLAPSE_EXPAND);
              EObjectFallbackSynthesis.this._kRenderingExtensions.to(EObjectFallbackSynthesis.this._kRenderingExtensions.from(EObjectFallbackSynthesis.this._kRenderingExtensions.setGridPlacementData(it), EObjectFallbackSynthesis.this._kRenderingExtensions.LEFT, 8, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.TOP, 0, 0), EObjectFallbackSynthesis.this._kRenderingExtensions.RIGHT, 8, 0, EObjectFallbackSynthesis.this._kRenderingExtensions.BOTTOM, 8, 0);
            }
          };
          ObjectExtensions.<KText>operator_doubleArrow(_addText, _function_1);
        }
      }
    };
    ObjectExtensions.<KRoundedRectangle>operator_doubleArrow(_createFigure_1, _function_1);
    this.nodeCache.put(object, node);
    return node;
  }
  
  /**
   * Translate a structural container feature into a port.
   */
  private KPort transformToPort(final EStructuralFeature containerFeature, final KNode node) {
    final KPort port = KGraphUtil.createInitializedPort();
    EList<KPort> _ports = node.getPorts();
    _ports.add(port);
    this._kPortExtensions.setPortSize(port, (this._kPortExtensions.getPortEdgeLength()).floatValue(), (this._kPortExtensions.getPortEdgeLength()).floatValue());
    this._kPortExtensions.<PortSide>addLayoutParam(port, CoreOptions.PORT_SIDE, PortSide.EAST);
    float _width = node.getWidth();
    float _minus = (_width - 1);
    this._kPortExtensions.setPortPos(port, _minus, (this._kPortExtensions.nextEPortYPosition(node)).floatValue());
    KLabel _createLabel = this._kLabelExtensions.createLabel(port);
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      @Override
      public void apply(final KLabel it) {
        it.setText(containerFeature.getName());
        final Bounds size = PlacementUtil.estimateSize(it);
        it.setWidth(size.getWidth());
        it.setHeight(size.getHeight());
      }
    };
    ObjectExtensions.<KLabel>operator_doubleArrow(_createLabel, _function);
    this.portCache.put(containerFeature, node, port);
    return port;
  }
  
  /**
   * Creates the visual representation of a node
   */
  private KRoundedRectangle createFigure(final KNode node) {
    final KRoundedRectangle figure = this._kRenderingExtensions.addRoundedRectangle(node, 8, 8, 1);
    this._kRenderingExtensions.setLineWidth(figure, 1);
    this._kRenderingExtensions.setForeground(figure, Colors.GRAY);
    this._kRenderingExtensions.setBackground(figure, Colors.GRAY_95);
    float _cornerWidth = figure.getCornerWidth();
    float _multiply = (2 * _cornerWidth);
    float _cornerHeight = figure.getCornerHeight();
    float _multiply_1 = (2 * _cornerHeight);
    this._kNodeExtensions.setMinimalNodeSize(node, _multiply, _multiply_1);
    return figure;
  }
}
